@@ -57,6 +57,56 @@ class CommonTsetlinMachine():
 		self.prepare_encode = mod_encode.get_function("prepare_encode")
 		self.encode = mod_encode.get_function("encode")
 
+	def __getstate__(self):
+		state = self.__dict__.copy()
+		state['ta_state'] = np.empty(self.number_of_classes*self.number_of_clauses*self.number_of_ta_chunks*self.number_of_state_bits).astype(np.uint32)
+		cuda.memcpy_dtoh(state['ta_state'], self.ta_state_gpu)
+		state['clause_weights'] = np.empty(self.number_of_classes*self.number_of_clauses).astype(np.uint8)
+		cuda.memcpy_dtoh(state['clause_weights'], self.clause_weights_gpu)
+		
+		if 'X_train' in state:
+			del state['X_train']
+			del state['Y_train']
+			del state['encoded_Y_train']
+			del state['encoded_X_training_gpu']
+			del state['update']
+
+		if 'X_test' in state:
+			del state['X_test']
+
+		if 'encoded_X_test_gpu' in state:
+			del state['encoded_X_test_gpu']
+
+		if 'evaluate' in state:
+			del state['evaluate']
+
+		if 'prepare_encode' in state:
+			del state['prepare_encode']
+			del state['prepare']
+			del state['encode']
+			del state['ta_state_gpu']
+			del state['clause_weights_gpu']
+			del state['clause_output_gpu']
+			del state['class_sum_gpu']
+			del state['Y_gpu']			
+
+		return state
+
+	def __setstate__(self, state):
+		self.__dict__.update(state)
+		self.X_train = np.array([])
+		self.Y_train = np.array([])
+		self.X_test = np.array([])
+
+		mod_encode = SourceModule(kernels.code_encode, no_extern_c=True)
+		self.prepare_encode = mod_encode.get_function("prepare_encode")
+		self.encode = mod_encode.get_function("encode")
+		
+		self.ta_state_gpu = cuda.mem_alloc(self.number_of_classes*self.number_of_clauses*self.number_of_ta_chunks*self.number_of_state_bits*4)
+		self.clause_weights_gpu = cuda.mem_alloc(self.number_of_classes*self.number_of_clauses)
+		cuda.memcpy_htod(self.ta_state_gpu, self.ta_state)
+		cuda.memcpy_htod(self.clause_weights_gpu, self.clause_weights)
+
 	def encode_X(self, X, encoded_X_gpu):
 		number_of_examples = X.shape[0]
 
@@ -88,39 +138,6 @@ class CommonTsetlinMachine():
 
 		return (ta_state[mc_tm_class, clause, ta // 32, self.number_of_state_bits-1] & (1 << (ta % 32))) > 0
 
-	def get_state(self):
-		if np.array_equal(self.clause_weights, np.array([])):
-			self.ta_state = np.empty(self.number_of_classes*self.number_of_clauses*self.number_of_ta_chunks*self.number_of_state_bits).astype(np.uint32)
-			cuda.memcpy_dtoh(self.ta_state, self.ta_state_gpu)
-			self.clause_weights = np.empty(self.number_of_classes*self.number_of_clauses).astype(np.uint8)
-			cuda.memcpy_dtoh(self.clause_weights, self.clause_weights_gpu)
-		return((self.ta_state, self.clause_weights, self.number_of_classes, self.number_of_clauses, self.number_of_features, self.dim, self.patch_dim, self.number_of_patches, self.number_of_state_bits, self.max_weight, self.number_of_ta_chunks, self.append_negated, self.min_y, self.max_y))
-
-	def set_state(self, state):
-		self.number_of_classes = state[2]
-		self.number_of_clauses = state[3]
-		self.number_of_features = state[4]
-		self.dim = state[5]
-		self.patch_dim = state[6]
-		self.number_of_patches = state[7]
-		self.number_of_state_bits = state[8]
-		self.max_weight = state[9]
-		self.number_of_ta_chunks = state[10]
-		self.append_negated = state[11]
-		self.min_y = state[12]
-		self.max_y = state[13]
-		
-		self.ta_state_gpu = cuda.mem_alloc(self.number_of_classes*self.number_of_clauses*self.number_of_ta_chunks*self.number_of_state_bits*4)
-		self.clause_weights_gpu = cuda.mem_alloc(self.number_of_classes*self.number_of_clauses)
-		cuda.memcpy_htod(self.ta_state_gpu, state[0])
-		cuda.memcpy_htod(self.clause_weights_gpu, state[1])
-
-		self.X_train = np.array([])
-		self.Y_train = np.array([])
-		self.X_test = np.array([])
-		self.ta_state = np.array([])
-		self.clause_weights = np.array([])
-
 	# Transform input data for processing at next layer
 	def transform(self, X):
 		number_of_examples = X.shape[0]
